
; 
; MD5 Procedures Mk 2
; 

; This code is stripped straight out of my DLL source code. My primary goal
; when writing this was to make the code work as stdcall DLL calls from the
; VB app I'm developing, so that I could pass simple parameters like the
; address of the byte array to hash, the length of it, the start point if
; only doing a part of a larger array (or file) etc., all error-checked
; where necessary/possible. That is why it is written in this format.

; My secondary goal was to allow fast access to these routines from within
; the DLL, as the MD5 routines in particular are used continually by PAR
; and PAR2 processing functions, all of which will be coded into the DLL in
; Assembler eventually. As a model for this style of compressed function
; call coding, allowing the stdcall routines to fall through into the
; syscall ones so that the parameters can be passed directly in registers,
; I used the great resource (and indispensible real-world example of MASM
; syntax DLL writing for which I'm extremely grateful!) at:
; http://www.geocities.com/SiliconValley/Heights/7394/

; Personally, as a beginning Assembler programmer, I find this explicit
; stack and parameter handling far easier to understand than using automatic
; prologues/epilogues, as you can see exactly what you're putting on the
; stack and why! It also means that I can make the most efficient use of
; register/stack storage, it makes this mixed-access approach easier and it
; codes smaller, so I'm sorry if it offends people with different ways of
; doing this. I'm sure you can all rewrite it better than me... ;-)

; An MD5 hash of a byte array requires three distinct stages: a call to set
; the initial values of the hash buffer, processing of the array from the
; start in exact multiples of 64 bytes, then processing of any sub-64-byte
; remainder. This gets padded to either 64 or 128 bytes depending on the
; size of the remainder. (This is because the byte array has its length
; in bits added to the end as part of the hash source, so the addition of
; 9 bytes might make a previously sub-64-byte remainder more than 64 bytes).

; In this code, the MD5Init call is self-explanatory. MD5ProcessByteArray
; can then be called any number of times with contiguous data but the
; buffer length MUST be an exact multiple of 64 bytes. The hash MUST be
; finished with a call to MD5ProcessFinal, which takes care of any remainder
; and handles the padding. As the hash value output is 4 DWORDS long, the
; MD5Init stdcall proc allows an external caller to provide a pointer to a
; 16-byte array for MD5ProcessFinal to fill the return value in.

; Because of the VB heritage of this code, the bufferlength parameters for
; the calls are single DWORD only (ie. max size 4Gb, or 2Gb using VB LONGs
; which are signed 32-bit integers, the largest size of a VB byte array or
; file length). These can easily be converted to full 64-bit if required.

; To manage the general case of a byte array and start offset being given
; that are not DWORD aligned in memory, which would severely affect the
; performance of the transform loop, the loop first copies any non-aligned
; data to the MD5Buffer in 64Kb chunks. Time of allocation of this buffer
; and its actual size is arbitrary - if you need it, I have added optional
; code in comments that would manage this transparently using API calls.

; Obviously the meat of this code is the transform loop, which operates on
; only 64 bytes of source data at a time. This is what requires the most
; optimisation, and I have spent much time perusing Agner Fog's resource
; and others to try and get a good feel for general optimisation without
; getting worked up over the exact number of uops or the exact processor
; it's optimised for (I assume P4) etc. I do, however, use the simplest,
; smallest and fastest instructions I can, suitable to my simple level
; of competence, and I know it's already pretty good... ;-)

; This is where I'd appreciate any pointers on correcting optimisation
; errors and code speed/size improvements. Many thanks in anticipation.

; IanB


; 

      .486                      ; create 32 bit code
      .model flat, stdcall      ; 32 bit memory model
      option casemap :none      ; case sensitive

      option prologue:none      ; no entry/exit code for subs
      option epilogue:none

; 

        MINUS_1                 EQU 0ffffffffH
        VB_TRUE                 EQU MINUS_1
        KBytes64                EQU 65536               ; you could use any buffer size you want
                                                        ; I use these multiples as they match the
                                                        ; API file I/O buffers exactly

        MD5STATE STRUCT
          A                     DWORD ?                 ; hash value is 128-bit
          B                     DWORD ?
          C                     DWORD ?
          D                     DWORD ?
          TotalLow              DWORD ?                 ; we need to save the amount of data hashed
          TotalHigh             DWORD ?                 ; for the end of the MD5 process
        MD5STATE ENDS

;       include <your correct path>\kernel32.inc        ; if you need the memory allocation functions

; 

      .data

        MD5Buffer               DWORD 0                 ; pointer to start of 64K MD5 buffer block

        MD5ResultPointer        DWORD 0                 ; pointer provided by external caller
                                                        ; must be zeroed initially

      .data?

        MD5State                MD5STATE (?)            ; storage for intermediate MD5 results

      .code


; 

align 64

MD5Init    proc    stdcall public, PointerToResultArray:DWORD

 comment    * -----------------------------------------------------------------------------
            The external calling procedure needs to provide a 16-byte array pointer to
            store the MD5 result. Local DLL code can use the result in MD5State directly,
            of course. We therefore zero out MD5ResultPointer after every use and test it
            for null before final output to see if the result is needed externally.

            If you are using the optional buffer allocation code, you should check for a
            non-zero return value in EAX to show that the allocation was successful.
            ----------------------------------------------------------------------------- *

        pop     eax                     ; save return address
        pop     edx                     ; PointerToResultArray
        push    eax                     ; restore return address
        mov     MD5ResultPointer, edx
                                        ; all operands are now cleared from stack
                                        ; fall through to syscall procedure

MD5Init    endp

align 4

MD5init    proc    syscall private

        ; On entry:
        ;       MD5ResultPointer has been set or is NULL if this is a syscall
        ; On exit:
        ;       EAX = 0 if unsuccessful buffer allocation, else non-zero
        ; EDX/EBX/EBP/EDI/ESI are preserved

        sub     eax, eax                ; EAX = 0

        mov     MD5State.A, 67452301H   ; set all start values
        mov     MD5State.B, 0efcdab89H
        mov     MD5State.TotalLow, eax
        mov     MD5State.TotalHigh, eax
        mov     MD5State.C, 98badcfeH
        mov     MD5State.D, 10325476H
        not     eax                     ; EAX = VB TRUE

 comment    * -----------------------------------------------------------------------------
            In the context of my DLL, I allocate all the buffers I need when it is first
            initialised and loaded. Here I've added some optional code that will allocate
            the process buffer if required. There is matching code in the MD5ProcessFinal
            function to release it and in both the process functions to check you've done
            the allocation. You may want to uncomment the following lines.

        invoke  VirtualAlloc, NULL, KBytes64, MEM_COMMIT, PAGE_READWRITE
        mov     MD5Buffer, eax
        ; VirtualAlloc returns pointer in EAX or zero if allocation unsuccessful
            ----------------------------------------------------------------------------- *

        ret

MD5init    endp

; 

align 4

MD5ProcessByteArray    proc    stdcall public, SourceArray:DWORD,SourceStart:DWORD,NumBytes:DWORD

 comment    * -----------------------------------------------------------------------------
            The number of bytes MUST be a multiple of 64, else use MD5ProcessFinal.
            MD5Init MUST have been called before this function.

            This function doesn't do the final message appends, so further arrays can be
            successively processed (thereby overcoming the 4Gb NumBytes limitation).
            Having separate Source and Start values allows calling from VB using ByRef to
            pass parts of byte arrays.
            ----------------------------------------------------------------------------- *

        pop     eax                     ; save return address
        pop     edx
        pop     ecx
        add     edx, ecx                ; pointer into SourceArray
        pop     ecx                     ; NumBytes
        push    eax                     ; restore return address
                                        ; all operands are now cleared from stack

        mov     eax, 3fH
        test    ecx, ecx
        jz      MD5ProcessError         ; nothing to process
        and     eax, ecx
        jnz     MD5ProcessError         ; not a multiple of 64 bytes
        test    edx, edx
        jz      MD5ProcessError         ; avoid NULL pointers
                                        ; fall through to syscall procedure

MD5ProcessByteArray    endp

align 4

MD5processbytearray    proc    syscall private

        ; On entry:
        ;       EDX = startpoint address of array including offset
        ;       ECX = number of bytes to process in this subarray
        ; On exit:
        ;       EAX = -1 if successful (or zero if parameters wrong)
        ;       ECX = 0 if successful
        ;       EDX = pointer to MD5State
        ; EBX/EBP/EDI/ESI preserved

 comment    * -----------------------------------------------------------------------------
            If you are doing your own buffer allocation, you may want to uncomment the
            following lines to check you've done it!

        mov     eax, MD5Buffer
        test    eax, eax
        jz      MD5ProcessError         ; avoid NULL pointers
            ----------------------------------------------------------------------------- *

        sub     eax, eax                ; EAX = 0
        add     MD5State.TotalLow, ecx  ; update bytes hashed total
        adc     MD5State.TotalHigh, eax ; any overflow of the high DWORD unimportant!

        push    ebp                     ; save registers needed for the Transform procedure
        push    ebx
        push    edi
        push    esi

        mov     eax, 3H                 ; check if the source is DWORD aligned
        mov     ebx, MD5State.A         ; get initial MD5State values for this call
        mov     ebp, MD5State.B
        mov     edi, MD5State.C
        mov     esi, MD5State.D
        and     eax, edx                ; EAX holds alignment offset

        call    MD5Transform

        or      eax, VB_TRUE            ; set successful return value = VB TRUE
        mov     MD5State.A, ebx         ; save intermediate MD5State values for next call
        mov     MD5State.B, ebp
        mov     MD5State.C, edi
        mov     MD5State.D, esi

        pop     esi                     ; restore registers
        pop     edi
        pop     ebx
        pop     ebp
        ret

MD5ProcessError:
        sub     eax, eax                ; zero result register
        ret

MD5processbytearray    endp

; 

align 4

MD5ProcessFinal    proc    stdcall public, SourceArray:DWORD,SourceStart:DWORD,InputLen:DWORD

 comment    * ------------------------------------------------------------------------------
            To finish, the message is "padded" (extended) so that its length (in bytes) is
            congruent to 56, modulo 64 (or 448/512 bits). Padding is always performed, even
            if the length of the message is already congruent to 56, modulo 64.

            Padding is performed as follows: a single "1" bit is appended to the message,
            and then "0" bits are appended until the length is congruent to 448, modulo 512.
            In all, at least one bit and at most 512 bits are appended.

            A 64-bit representation of the bit-length of the message before the padding bits
            were added is then appended to the result of the previous step. In the unlikely
            event that is greater than 2^64, then only the low-order 64 bits are used.
            (These bits are appended as two 32-bit words and appended low-order word first.)

            (Above description quoted from the original MD5 specification document.)

            This function returns VB TRUE (-1) in EAX if successful, and if MD5ResultPointer
            is not NULL then it copies MD5State to that array for the external calling app.
            Having separate Source and Start values allows calling from VB using ByRef to
            pass parts of byte arrays. MD5Init MUST have been called before this function.
            ------------------------------------------------------------------------------ *

        pop     eax                     ; save return address
        pop     edx
        pop     ecx
        add     edx, ecx                ; pointer into SourceArray
        pop     ecx                     ; InputLen
        push    eax                     ; restore return address
                                        ; all operands are now cleared from stack

        sub     eax, eax                ; zero result register
        test    ecx, ecx
        jz      EndMD5FinalProcess      ; nothing to process
        test    edx, edx
        jz      EndMD5FinalProcess      ; avoid NULL pointers
                                        ; fall through to syscall procedure

MD5ProcessFinal    endp

align 4

MD5processfinal    proc    syscall private

        ; On entry:
        ;       EDX = startpoint address of array including offset
        ;       ECX = number of bytes to process
        ; On exit:
        ;       EAX = -1 = VB TRUE (or zero if parameters wrong)
        ;       4 DWORDS at a valid MD5ResultPointer address has MD5State copied into it
        ; EBX/EBP/EDI/ESI preserved

 comment    * -----------------------------------------------------------------------------
            If you are doing your own buffer allocation, you may want to uncomment the
            following lines to check you've done it!

        mov     eax, MD5Buffer
        test    eax, eax
        jz      MD5ProcessError         ; avoid NULL pointers
            ----------------------------------------------------------------------------- *

        sub     eax, eax                ; EAX = 0
        add     MD5State.TotalLow, ecx  ; update bytes hashed total
        adc     MD5State.TotalHigh, eax ; any overflow of the high DWORD unimportant!

        push    ebx                     ; save registers needed for the Transform procedure
        push    ebp
        push    edi
        push    esi

        add     eax, 3fH
        mov     ebx, MD5State.A
        and     eax, ecx                ; get remainder modulo 64
        mov     ebp, MD5State.B
        push    eax                     ; save remainder bytes number
        mov     edi, MD5State.C
        mov     esi, MD5State.D
        mov     eax, 3H                 ; check if the source is DWORD aligned
        and     ecx, 0ffffffc0H         ; keep multiple of 64 bytes left as counter
        jz      @F

        and     eax, edx                ; EAX holds alignment offset
        call    MD5Transform            ; keep all MD5State values in registers between calls
                                        ; this hashes all but the last sub-64 byte remainder of the array
@@:
        pop     ecx                     ; retrieve remainder bytes number
        push    ebx                     ; save MD5State.A to free up extra register
        mov     eax, 128                ; max amount of remainder + padding
        mov     ebx, edx                ; EBX now holds pointer to any remainder bytes in SourceArray
        mov     edx, MD5Buffer          ; EDX now holds pointer to MD5Buffer as our new source array
                                        ; we will copy all our remainder data and padding there

BufferEraseLoop:
        sub     eax, 4
        jc      @F                      ; EAX must be zero or more
        mov     DWORD PTR [edx+eax], 0
        cmp     eax, ecx                ; zero out MD5Buffer from top until EAX < remainder
        jnb     BufferEraseLoop
@@:
        mov     BYTE PTR [edx+ecx], 80h ; append the single "1" bit after any remainder
        test    ecx, ecx
        jz      TransformFinal          ; skip byte transfer if there is no remainder
        mov     eax, ecx                ; save remainder counter - byte value is in AL
        sub     ecx, 1                  ; remainder is 1-based, array pointer is 0-based

TransferRemainderLoop:
        mov     ah, [ebx+ecx]           ; put remainder of bytes to hash into the buffer
        mov     [edx+ecx], ah           ; a byte at a time
        sub     ecx, 1
        jnc     TransferRemainderLoop   ; if no loop back, ECX = -1

        cmp     al, 56                  ; if less than 8 bytes spare, more padding is needed
        adc     ecx, 0                  ; so we use the top 64 bytes of MD5Buffer as well
        and     ecx, 64                 ; this code avoids a conditional jump to add 64 to ECX

TransformFinal:
        push    ebp                     ; save MD5State.B to free up extra register
        mov     eax, MD5State.TotalLow
        mov     ebx, MD5State.TotalHigh
        rol     eax, 3                  ; multiply byte count by 8 for bits, keeping bottom 3
        shl     ebx, 3                  ; multiply by 8 for bits, losing any in top 3
        mov     ebp, eax                ; copy value
        and     eax, 0fffffff8H         ; mask out the top 3 bits we rotated round
        and     ebp, 7H                 ; mask out the rest of the bit total
        mov     [edx+ecx+56], eax
        or      ebx, ebp                ; combine with previous shifted 3 bits
        mov     [edx+ecx+60], ebx

        add     ecx, 64                 ; we will finally process either 64 or 128 bytes
        pop     ebp                     ; retrieve MD5State.B
        pop     ebx                     ; retrieve MD5State.A

        sub     eax, eax                ; MD5Buffer is DWORD aligned
        call    MD5Transform            ; final MD5State values returned in registers

        mov     ecx, MD5ResultPointer
        mov     MD5State.A, ebx         ; save final values
        mov     MD5State.B, ebp
        mov     MD5State.C, edi
        mov     MD5State.D, esi

        test    ecx, ecx                ; skip if it's a null pointer
        jz      @F
        mov     [ecx], ebx              ; otherwise copy final values to provided external array
        mov     [ecx+4], ebp
        mov     [ecx+8], edi
        mov     [ecx+12], esi
        mov     DWORD PTR MD5ResultPointer, 0       ; reset the pointer as NULL
@@:

 comment    * -----------------------------------------------------------------------------
            If you need to undo your own buffer allocation, you may want to uncomment the
            following lines.

        mov     eax, MD5Buffer
        invoke  VirtualFree, eax, NULL, MEM_RELEASE
        mov     DWORD PTR MD5Buffer, 0
            ----------------------------------------------------------------------------- *

        or      eax, VB_TRUE            ; set successful return value = VB TRUE
        pop     esi                     ; restore registers
        pop     edi
        pop     ebp
        pop     ebx

EndMD5FinalProcess:
        ret

MD5processfinal    endp

; 

align 4

MD5Transform    proc    syscall private

 comment    * ------------------------------------------------------------------------------
            This processes a byte array sequentially in 64 byte blocks.

            On entry:
                EBX/EBP/EDI/ESI = current MD5State values (= A, B, C, D)
                ECX = loop counter, multiple of 64
                EDX = pointer to first 64-byte block in the source array
                EAX = DWORD alignment of EDX pointer = value of lowest 2 bits
            On exit:
                EBX/EBP/EDI/ESI = new MD5State values (= A, B, C, D)
                ECX = 0
                EDX = pointer to next 64-byte block to process, may not be valid
            No registers are preserved as it is called from the context of other functions.

            The hash transformation is of this form where X is a DWORD from the block and
            T is a different constant for each iteration:

                For each 64-byte data block, using the previous accumulated value to start:
                    Repeat for 4 different functions:
                        Repeat function 16 times:
                            A = A + f(B, C, D) + X(n) + T
                            Rotate A left by S bits
                            A = A + B
                            Rotate A -> B -> C -> D -> A
                        Next iteration
                    Next function
                Add new results to starting value

            We can partially pair successive iterations to process in parallel and avoid
            register access stalls. This is also why the block pointer EDX is sometimes
            explicitly updated, to insert an extra instruction in that iteration, rather
            than always using [EDX+offset] addressing.

            EDX is preferred to EBP for this as according to Fog Chapter 6: "Instructions
            with EBP as base pointer and no displacement and no index take one byte more
            than with other registers". ADD REG, [EDX] should therefore code smaller.
            ------------------------------------------------------------------------------ *

        test    eax, eax            ; test array pointer is starting on a DWORD boundary
        jnz     CopyTransformData
        push    eax                 ; save alignment value for end of loop as flag

MD5TransformLoop:
        push    ecx                 ; save ECX as we need another spare register
        push    esi                 ; save MD5State values at start for addition at end of loop
        push    edi
        push    ebp
        push    ebx

 comment    * -------------------------------------------------------------------
            F function:
                A = A + ((B And C) Or (Not (B) And D)) + X(n) + T
                or alternatively A = A + (((D Xor C) And B) Xor D) + X(n) + T
                Rotate A left by S bits (7, 12, 17, 22, used in rotation)
                A = A + B
                Rotate registers around
            ------------------------------------------------------------------- *

        mov     eax, esi            ; A in EBX, B in EBP, C in EDI, D in ESI
        mov     ecx, edi            ; set ECX now for next iteration
        xor     eax, edi
        xor     ecx, ebp
        and     eax, ebp
        add     ebx, 0d76aa478H     ; add T
        xor     eax, esi
        add     ebx, [edx]          ; add X(0)
        add     esi, 0e8c7b756H     ; add T to next A
        add     ebx, eax            ; add function result
        add     edx, 4
        rol     ebx, 7              ; rotate
        add     esi, [edx]          ; add X(1) to next A
        add     ebx, ebp            ; add B

        mov     eax, ebp            ; set EAX now for next iteration
        and     ecx, ebx            ; A in ESI, B in EBX, C in EBP, D in EDI
        xor     eax, ebx
        xor     ecx, edi
        add     edx, 4
        add     esi, ecx            ; add function result
        add     edi, 242070dbH      ; add T to next A
        rol     esi, 12             ; rotate
        add     edi, [edx]          ; add X(2) to next A
        add     esi, ebx            ; add B

        mov     ecx, ebx            ; set ECX now for next iteration
        and     eax, esi            ; A in EDI, B in ESI, C in EBX, D in EBP
        xor     ecx, esi
        xor     eax, ebp
        add     edx, 4
        add     edi, eax            ; add function result
        add     ebp, 0c1bdceeeH     ; add T to next A
        rol     edi, 17             ; rotate
        add     ebp, [edx]          ; add X(3) to next A
        add     edi, esi            ; add B

        mov     eax, esi            ; set EAX now for next iteration
        and     ecx, edi            ; A in EBP, B in EDI, C in ESI, D in EBX
        xor     eax, edi
        xor     ecx, ebx
        add     edx, 4
        add     ebp, ecx            ; add function result
        add     ebx, 0f57c0fafH     ; add T to next A
        rol     ebp, 22             ; rotate
        add     ebx, [edx]          ; add X(4) to next A
        add     ebp, edi            ; add B

        mov     ecx, edi            ; set ECX now for next iteration
        and     eax, ebp            ; A in EBX, B in EBP, C in EDI, D in ESI
        xor     ecx, ebp
        xor     eax, esi
        add     edx, 4
        add     ebx, eax            ; add function result
        add     esi, 4787c62aH      ; add T to next A
        rol     ebx, 7              ; rotate
        add     esi, [edx]          ; add X(5) to next A
        add     ebx, ebp            ; add B

        mov     eax, ebp            ; set EAX now for next iteration
        and     ecx, ebx            ; A in ESI, B in EBX, C in EBP, D in EDI
        xor     eax, ebx
        xor     ecx, edi
        add     edx, 4
        add     esi, ecx            ; add function result
        add     edi, 0a8304613H     ; add T to next A
        rol     esi, 12             ; rotate
        add     edi, [edx]          ; add X(6) to next A
        add     esi, ebx            ; add B

        mov     ecx, ebx            ; set ECX now for next iteration
        and     eax, esi            ; A in EDI, B in ESI, C in EBX, D in EBP
        xor     ecx, esi
        xor     eax, ebp
        add     edx, 4
        add     edi, eax            ; add function result
        add     ebp, 0fd469501H     ; add T to next A
        rol     edi, 17             ; rotate
        add     ebp, [edx]          ; add X(7) to next A
        add     edi, esi            ; add B

        mov     eax, esi            ; set EAX now for next iteration
        and     ecx, edi            ; A in EBP, B in EDI, C in ESI, D in EBX
        xor     eax, edi
        xor     ecx, ebx
        add     edx, 4
        add     ebp, ecx            ; add function result
        add     ebx, 698098d8H      ; add T to next A
        rol     ebp, 22             ; rotate
        add     ebx, [edx]          ; add X(8) to next A
        add     ebp, edi            ; add B

        mov     ecx, edi            ; set ECX now for next iteration
        and     eax, ebp            ; A in EBX, B in EBP, C in EDI, D in ESI
        xor     ecx, ebp
        xor     eax, esi
        add     edx, 4
        add     ebx, eax            ; add function result
        add     esi, 8b44f7afH      ; add T to next A
        rol     ebx, 7              ; rotate
        add     esi, [edx]          ; add X(9) to next A
        add     ebx, ebp            ; add B

        mov     eax, ebp            ; set EAX now for next iteration
        and     ecx, ebx            ; A in ESI, B in EBX, C in EBP, D in EDI
        xor     eax, ebx
        xor     ecx, edi
        add     edx, 4
        add     esi, ecx            ; add function result
        add     edi, 0ffff5bb1H     ; add T to next A
        rol     esi, 12             ; rotate
        add     edi, [edx]          ; add X(10) to next A
        add     esi, ebx            ; add B

        mov     ecx, ebx            ; set ECX now for next iteration
        and     eax, esi            ; A in EDI, B in ESI, C in EBX, D in EBP
        xor     ecx, esi
        xor     eax, ebp
        add     edx, 4
        add     edi, eax            ; add function result
        add     ebp, 895cd7beH      ; add T to next A
        rol     edi, 17             ; rotate
        add     ebp, [edx]          ; add X(11) to next A
        add     edi, esi            ; add B

        mov     eax, esi            ; set EAX now for next iteration
        and     ecx, edi            ; A in EBP, B in EDI, C in ESI, D in EBX
        xor     eax, edi
        xor     ecx, ebx
        add     edx, 4
        add     ebp, ecx            ; add function result
        add     ebx, 6b901122H      ; add T to next A
        rol     ebp, 22             ; rotate
        add     ebx, [edx]          ; add X(12) to next A
        add     ebp, edi            ; add B

        mov     ecx, edi            ; set ECX now for next iteration
        and     eax, ebp            ; A in EBX, B in EBP, C in EDI, D in ESI
        xor     ecx, ebp
        xor     eax, esi
        add     edx, 4
        add     ebx, eax            ; add function result
        add     esi, 0fd987193H     ; add T to next A
        rol     ebx, 7              ; rotate
        add     esi, [edx]          ; add X(13) to next A
        add     ebx, ebp            ; add B

        mov     eax, ebp            ; set EAX now for next iteration
        and     ecx, ebx            ; A in ESI, B in EBX, C in EBP, D in EDI
        xor     eax, ebx
        xor     ecx, edi
        add     edx, 4
        add     esi, ecx            ; add function result
        add     edi, 0a679438eH     ; add T to next A
        rol     esi, 12             ; rotate
        add     edi, [edx]          ; add X(14) to next A
        add     esi, ebx            ; add B

        mov     ecx, ebx            ; set ECX now for next iteration
        and     eax, esi            ; A in EDI, B in ESI, C in EBX, D in EBP
        xor     ecx, esi
        xor     eax, ebp
        add     edx, 4
        add     edi, eax            ; add function result
        add     ebp, 49b40821H      ; add T to next A
        rol     edi, 17             ; rotate
        add     ebp, [edx]          ; add X(15) to next A
        add     edi, esi            ; add B

        mov     eax, esi            ; set EAX now for next iteration
        and     ecx, edi            ; A in EBP, B in EDI, C in ESI, D in EBX
        not     eax
        xor     ecx, ebx
        sub     edx, 60             ; EDX returned to base value
        add     ebp, ecx            ; add function result
        add     ebx, 0f61e2562H     ; add T to next A
        rol     ebp, 22             ; rotate
        add     ebx, [edx+4]        ; add X(1) to next A
        add     ebp, edi            ; add B

 comment    * -------------------------------------------------------------------
            G Function:
                A = A + ((B And D) Or (C And Not (D))) + X(n) + T
                Rotate A left by S bits (5, 9, 14, 20, used in rotation)
                A = A + B
                Rotate registers around
            ------------------------------------------------------------------- *

        mov     ecx, esi            ; A in EBX, B in EBP, C in EDI, D in ESI
        and     eax, edi
        and     ecx, ebp
        add     esi, 0c040b340H     ; add T to next A
        or      eax, ecx
        mov     ecx, edi            ; set ECX now for next iteration
        add     ebx, eax            ; add function result
        not     ecx
        rol     ebx, 5              ; rotate
        add     esi, [edx+24]       ; add X(6) to next A
        add     ebx, ebp            ; add B

        mov     eax, edi            ; A in ESI, B in EBX, C in EBP, D in EDI
        and     ecx, ebp
        and     eax, ebx
        add     edi, 265e5a51H      ; add T to next A
        or      eax, ecx
        mov     ecx, ebp            ; set ECX now for next iteration
        add     esi, eax            ; add function result
        not     ecx
        rol     esi, 9              ; rotate
        add     edi, [edx+44]       ; add X(11) to next A
        add     esi, ebx            ; add B

        mov     eax, ebp            ; A in EDI, B in ESI, C in EBX, D in EBP
        and     ecx, ebx
        and     eax, esi
        add     ebp, 0e9b6c7aaH     ; add T to next A
        or      eax, ecx
        mov     ecx, ebx            ; set ECX now for next iteration
        add     edi, eax            ; add function result
        not     ecx
        rol     edi, 14             ; rotate
        add     ebp, [edx]          ; add X(0) to next A
        add     edi, esi            ; add B

        mov     eax, ebx            ; A in EBP, B in EDI, C in ESI, D in EBX
        and     ecx, esi
        and     eax, edi
        add     ebx, 0d62f105dH     ; add T to next A
        or      eax, ecx
        mov     ecx, esi            ; set ECX now for next iteration
        add     ebp, eax            ; add function result
        not     ecx
        rol     ebp, 20             ; rotate
        add     ebx, [edx+20]       ; add X(5) to next A
        add     ebp, edi            ; add B

        mov     eax, esi            ; A in EBX, B in EBP, C in EDI, D in ESI
        and     ecx, edi
        and     eax, ebp
        add     esi, 2441453H       ; add T to next A
        or      eax, ecx
        mov     ecx, edi            ; set ECX now for next iteration
        add     ebx, eax            ; add function result
        not     ecx
        rol     ebx, 5              ; rotate
        add     esi, [edx+40]       ; add X(10) to next A
        add     ebx, ebp            ; add B

        mov     eax, edi            ; A in ESI, B in EBX, C in EBP, D in EDI
        and     ecx, ebp
        and     eax, ebx
        add     edi, 0d8a1e681H     ; add T to next A
        or      eax, ecx
        mov     ecx, ebp            ; set ECX now for next iteration
        add     esi, eax            ; add function result
        not     ecx
        rol     esi, 9              ; rotate
        add     edi, [edx+60]       ; add X(15) to next A
        add     esi, ebx            ; add B

        mov     eax, ebp            ; A in EDI, B in ESI, C in EBX, D in EBP
        and     ecx, ebx
        and     eax, esi
        add     ebp, 0e7d3fbc8H     ; add T to next A
        or      eax, ecx
        mov     ecx, ebx            ; set ECX now for next iteration
        add     edi, eax            ; add function result
        not     ecx
        rol     edi, 14             ; rotate
        add     ebp, [edx+16]       ; add X(4) to next A
        add     edi, esi            ; add B

        mov     eax, ebx            ; A in EBP, B in EDI, C in ESI, D in EBX
        and     ecx, esi
        and     eax, edi
        add     ebx, 21e1cde6H      ; add T to next A
        or      eax, ecx
        mov     ecx, esi            ; set ECX now for next iteration
        add     ebp, eax            ; add function result
        not     ecx
        rol     ebp, 20             ; rotate
        add     ebx, [edx+36]       ; add X(9) to next A
        add     ebp, edi            ; add B

        mov     eax, esi            ; A in EBX, B in EBP, C in EDI, D in ESI
        and     ecx, edi
        and     eax, edx
        add     esi, 0c33707d6H     ; add T to next A
        or      eax, ecx
        mov     ecx, edi            ; set ECX now for next iteration
        add     ebx, eax            ; add function result
        not     ecx
        rol     ebx, 5              ; rotate
        add     esi, [edx+56]       ; add X(14) to next A
        add     ebx, ebp            ; add B

        mov     eax, edi            ; A in ESI, B in EBX, C in EBP, D in EDI
        and     ecx, ebp
        and     eax, ebx
        add     edi, 0f4d50d87H     ; add T to next A
        or      eax, ecx
        mov     ecx, ebp            ; set ECX now for next iteration
        add     esi, eax            ; add function result
        not     ecx
        rol     esi, 9              ; rotate
        add     edi, [edx+12]       ; add X(3) to next A
        add     esi, ebx            ; add B

        mov     eax, ebp            ; A in EDI, B in ESI, C in EBX, D in EBP
        and     ecx, ebx
        and     eax, esi
        add     ebp, 455a14edH      ; add T to next A
        or      eax, ecx
        mov     ecx, ebx            ; set ECX now for next iteration
        add     edi, eax            ; add function result
        not     ecx
        rol     edi, 14             ; rotate
        add     ebp, [edx+32]       ; add X(8) to next A
        add     edi, esi            ; add B

        mov     eax, ebx            ; A in EBP, B in EDI, C in ESI, D in EBX
        and     ecx, esi
        and     eax, edi
        add     ebx, 0a9e3e905H     ; add T to next A
        or      eax, ecx
        mov     ecx, esi            ; set ECX now for next iteration
        add     ebp, eax            ; add function result
        not     ecx
        rol     ebp, 20             ; rotate
        add     ebx, [edx+52]       ; add X(13) to next A
        add     ebp, edi            ; add B

        mov     eax, esi            ; A in EBX, B in EBP, C in EDI, D in ESI
        and     ecx, edi
        and     eax, ebp
        add     esi, 0fcefa3f8H     ; add T to next A
        or      eax, ecx
        mov     ecx, edi            ; set ECX now for next iteration
        add     ebx, eax            ; add function result
        not     ecx
        rol     ebx, 5              ; rotate
        add     esi, [edx+8]        ; add X(2) to next A
        add     ebx, ebp            ; add B

        mov     eax, edi            ; A in ESI, B in EBX, C in EBP, D in EDI
        and     ecx, ebp
        and     eax, ebx
        add     edi, 676f02d9H      ; add T to next A
        or      eax, ecx
        mov     ecx, ebp            ; set ECX now for next iteration
        add     esi, eax            ; add function result
        not     ecx
        rol     esi, 9              ; rotate
        add     edi, [edx+28]       ; add X(7) to next A
        add     esi, ebx            ; add B

        mov     eax, ebp            ; A in EDI, B in ESI, C in EBX, D in EBP
        and     ecx, ebx
        and     eax, esi
        add     ebp, 8d2a4c8aH      ; add T to next A
        or      eax, ecx
        mov     ecx, ebx            ; set ECX now for next iteration
        add     edi, eax            ; add function result
        not     ecx
        rol     edi, 14             ; rotate
        add     ebp, [edx+48]       ; add X(12) to next A
        add     edi, esi            ; add B

        mov     eax, ebx            ; A in EBP, B in EDI, C in ESI, D in EBX
        and     ecx, esi
        and     eax, edi
        add     ebx, 0fffa3942H     ; add T to next A
        or      eax, ecx
        add     ebx, [edx+20]       ; add X(5) to next A
        add     ebp, eax            ; add function result
        mov     eax, esi            ; set EAX now for next iteration
        rol     ebp, 20             ; rotate
        xor     eax, edi
        add     ebp, edi            ; add B

 comment    * -------------------------------------------------------------------
            H function:
                A = A + (B Xor C Xor D) + X(n) + T
                Rotate A left by S bits (4, 11, 16, 23, used in rotation)
                A = A + B
                Rotate registers around
            ------------------------------------------------------------------- *

        mov     ecx, edi            ; set ECX now for next iteration
        xor     eax, ebp            ; A in EBX, B in Ebp, C in EDI, D in ESI
        xor     ecx, ebp
        add     ebx, eax            ; add function result
        add     esi, 8771f681H      ; add T to next A
        rol     ebx, 4              ; rotate
        add     esi, [edx+32]       ; add X(8) to next A
        add     ebx, ebp            ; add B

        mov     eax, ebp            ; set EAX now for next iteration
        xor     ecx, ebx            ; A in ESI, B in EBX, C in EBP, D in EDI
        xor     eax, ebx
        add     esi, ecx            ; add function result
        add     edi, 6d9d6122H      ; add T to next A
        rol     esi, 11             ; rotate
        add     edi, [edx+44]       ; add X(11) to next A
        add     esi, ebx            ; add B

        mov     ecx, ebx            ; set ECX now for next iteration
        xor     eax, esi            ; A in EDI, B in ESI, C in EBX, D in EBP
        xor     ecx, esi
        add     edi, eax            ; add function result
        add     ebp, 0fde5380cH     ; add T to next A
        rol     edi, 16             ; rotate
        add     ebp, [edx+56]       ; add X(14) to next A
        add     edi, esi            ; add B

        mov     eax, esi            ; set EAX now for next iteration
        xor     ecx, edi            ; A in EBP, B in EDI, C in ESI, D in EBX
        xor     eax, edi
        add     ebp, ecx            ; add function result
        add     ebx, 0a4beea44H     ; add T to next A
        rol     ebp, 23             ; rotate
        add     ebx, [edx+4]        ; add X(1) to next A
        add     ebp, edi            ; add B

        mov     ecx, edi            ; set ECX now for next iteration
        xor     eax, ebp            ; A in EBX, B in EBP, C in EDI, D in ESI
        xor     ecx, ebp
        add     ebx, eax            ; add function result
        add     esi, 4bdecfa9H      ; add T to next A
        rol     ebx, 4              ; rotate
        add     esi, [edx+16]       ; add X(4) to next A
        add     ebx, ebp            ; add B

        mov     eax, ebp            ; set EAX now for next iteration
        xor     ecx, ebx            ; A in ESI, B in EBX, C in EBP, D in EDI
        xor     eax, ebx
        add     esi, ecx            ; add function result
        add     edi, 0f6bb4b60H     ; add T to next A
        rol     esi, 11             ; rotate
        add     edi, [edx+28]       ; add X(7) to next A
        add     esi, ebx            ; add B

        mov     ecx, ebx            ; set ECX now for next iteration
        xor     eax, esi            ; A in EDI, B in ESI, C in EBX, D in EBP
        xor     ecx, esi
        add     edi, eax            ; add function result
        add     ebp, 0bebfbc70H     ; add T to next A
        rol     edi, 16             ; rotate
        add     ebp, [edx+40]       ; add X(10) to next A
        add     edi, esi            ; add B

        mov     eax, esi            ; set EAX now for next iteration
        xor     ecx, edi            ; A in EBP, B in EDI, C in ESI, D in EBX
        xor     eax, edi
        add     ebp, ecx            ; add function result
        add     ebx, 289b7ec6H      ; add T to next A
        rol     ebp, 23             ; rotate
        add     ebx, [edx+52]       ; add X(13) to next A
        add     ebp, edi            ; add B

        mov     ecx, edi            ; set ECX now for next iteration
        xor     eax, ebp            ; A in EBX, B in EBP, C in EDI, D in ESI
        xor     ecx, ebp
        add     ebx, eax            ; add function result
        add     esi, 0eaa127faH     ; add T to next A
        rol     ebx, 4              ; rotate
        add     esi, [edx]          ; add X(0) to next A
        add     ebx, ebp            ; add B

        mov     eax, ebp            ; set EAX now for next iteration
        xor     ecx, ebx            ; A in ESI, B in EBX, C in EBP, D in EDI
        xor     eax, ebx
        add     esi, ecx            ; add function result
        add     edi, 0d4ef3085H     ; add T to next A
        rol     esi, 11             ; rotate
        add     edi, [edx+12]       ; add X(3) to next A
        add     esi, ebx            ; add B

        mov     ecx, ebx            ; set ECX now for next iteration
        xor     eax, esi            ; A in EDI, B in ESI, C in EBX, D in EBP
        xor     ecx, esi
        add     edi, eax            ; add function result
        add     ebp, 4881d05H       ; add T to next A
        rol     edi, 16             ; rotate
        add     ebp, [edx+24]       ; add X(6) to next A
        add     edi, esi            ; add B

        mov     eax, esi            ; set EAX now for next iteration
        xor     ecx, edi            ; A in EBP, B in EDI, C in ESI, D in EBX
        xor     eax, edi
        add     ebp, ecx            ; add function result
        add     ebx, 0d9d4d039H     ; add T to next A
        rol     ebp, 23             ; rotate
        add     ebx, [edx+36]       ; add X(9) to next A
        add     ebp, edi            ; add B

        mov     ecx, edi            ; set ECX now for next iteration
        xor     eax, ebp            ; A in EBX, B in EBP, C in EDI, D in ESI
        xor     ecx, ebp
        add     ebx, eax            ; add function result
        add     esi, 0e6db99e5H     ; add T to next A
        rol     ebx, 4              ; rotate
        add     esi, [edx+48]       ; add X(12) to next A
        add     ebx, ebp            ; add B

        mov     eax, ebp            ; set EAX now for next iteration
        xor     ecx, ebx            ; A in ESI, B in EBX, C in EBP, D in EDI
        xor     eax, ebx
        add     esi, ecx            ; add function result
        add     edi, 1fa27cf8H      ; add T to next A
        rol     esi, 11             ; rotate
        add     edi, [edx+60]       ; add X(15) to next A
        add     esi, ebx            ; add B

        mov     ecx, ebx            ; set ECX now for next iteration
        xor     eax, esi            ; A in EDI, B in ESI, C in EBX, D in EBP
        xor     ecx, esi
        add     edi, eax            ; add function result
        add     ebp, 0c4ac5665H     ; add T to next A
        rol     edi, 16             ; rotate
        add     ebp, [edx+8]        ; add X(2) to next A
        add     edi, esi            ; add B

        mov     eax, esi            ; set EAX now for next iteration
        xor     ecx, edi            ; A in EBP, B in EDI, C in ESI, D in EBX
        not     eax
        add     ebp, ecx            ; add function result
        add     ebx, 0f4292244H     ; add T to next A
        rol     ebp, 23             ; rotate
        add     ebx, [edx]          ; add X(0) to next A
        add     ebp, edi            ; add B

 comment    * -------------------------------------------------------------------
            I function:
                A = A + (C Xor (B Or Not (D))) + X(n) + T
                Rotate A left by S bits (6, 10, 15, 21, used in rotation)
                A = A + B
                Rotate registers around
            ------------------------------------------------------------------- *

        mov     ecx, edi            ; set ECX now for next iteration
        or      eax, ebp            ; A in EBX, B in EBP, C in EDI, D in ESI
        not     ecx
        xor     eax, edi
        add     edx, 28
        add     ebx, eax            ; add function result
        add     esi, 432aff97H      ; add T to next A
        rol     ebx, 6              ; rotate
        add     esi, [edx]          ; add X(7) to next A
        add     ebx, ebp            ; add B

        mov     eax, ebp            ; set EAX now for next iteration
        or      ecx, ebx            ; A in ESI, B in EBX, C in EBP, D in EDI
        not     eax
        xor     ecx, ebp
        add     edx, 28
        add     esi, ecx            ; add function result
        add     edi, 0ab9423a7H     ; add T to next A
        rol     esi, 10             ; rotate
        add     edi, [edx]          ; add X(14) to next A
        add     esi, ebx            ; add B

        mov     ecx, ebx            ; set ECX now for next iteration
        or      eax, esi            ; A in EDI, B in ESI, C in EBX, D in EBP
        not     ecx
        xor     eax, ebx
        sub     edx, 36
        add     edi, eax            ; add function result
        add     ebp, 0fc93a039H     ; add T to next A
        rol     edi, 15             ; rotate
        add     ebp, [edx]          ; add X(5) to next A
        add     edi, esi            ; add B

        mov     eax, esi            ; set EAX now for next iteration
        or      ecx, edi            ; A in EBP, B in EDI, C in ESI, D in EBX
        not     eax
        xor     ecx, esi
        add     edx, 28
        add     ebp, ecx            ; add function result
        add     ebx, 655b59c3H      ; add T to next A
        rol     ebp, 21             ; rotate
        add     ebx, [edx]          ; add X(12) to next A
        add     ebp, edi            ; add B

        mov     ecx, edi            ; set ECX now for next iteration
        or      eax, ebp            ; A in EBX, B in EBP, C in EDI, D in ESI
        not     ecx
        xor     eax, edi
        sub     edx, 36
        add     ebx, eax            ; add function result
        add     esi, 8f0ccc92H      ; add T to next A
        rol     ebx, 6              ; rotate
        add     esi, [edx]          ; add X(3) to next A
        add     ebx, ebp            ; add B

        mov     eax, ebp            ; set EAX now for next iteration
        or      ecx, ebx            ; A in ESI, B in EBX, C in EBP, D in EDI
        not     eax
        xor     ecx, ebp
        add     edx, 28
        add     esi, ecx            ; add function result
        add     edi, 0ffeff47dH     ; add T to next A
        rol     esi, 10             ; rotate
        add     edi, [edx]          ; add X(10) to next A
        add     esi, ebx            ; add B

        mov     ecx, ebx            ; set ECX now for next iteration
        or      eax, esi            ; A in EDI, B in ESI, C in EBX, D in EBP
        not     ecx
        xor     eax, ebx
        sub     edx, 36
        add     edi, eax            ; add function result
        add     ebp, 85845dd1H      ; add T to next A
        rol     edi, 15             ; rotate
        add     ebp, [edx]          ; add X(1) to next A
        add     edi, esi            ; add B

        mov     eax, esi            ; set EAX now for next iteration
        or      ecx, edi            ; A in EBP, B in EDI, C in ESI, D in EBX
        not     eax
        xor     ecx, esi
        add     edx, 28
        add     ebp, ecx            ; add function result
        add     ebx, 6fa87e4fH      ; add T to next A
        rol     ebp, 21             ; rotate
        add     ebx, [edx]          ; add X(8) to next A
        add     ebp, edi            ; add B

        mov     ecx, edi            ; set ECX now for next iteration
        or      eax, ebp            ; A in EBX, B in EBP, C in EDI, D in ESI
        not     ecx
        xor     eax, edi
        add     edx, 28
        add     ebx, eax            ; add function result
        add     esi, 0fe2ce6e0H     ; add T to next A
        rol     ebx, 6              ; rotate
        add     esi, [edx]          ; add X(15) to next A
        add     ebx, ebp            ; add B

        mov     eax, ebp            ; set EAX now for next iteration
        or      ecx, ebx            ; A in ESI, B in EBX, C in EBP, D in EDI
        not     eax
        xor     ecx, ebp
        sub     edx, 36
        add     esi, ecx            ; add function result
        add     edi, 0a3014314H     ; add T to next A
        rol     esi, 10             ; rotate
        add     edi, [edx]          ; add X(6) to next A
        add     esi, ebx            ; add B

        mov     ecx, ebx            ; set ECX now for next iteration
        or      eax, esi            ; A in EDI, B in ESI, C in EBX, D in EBP
        not     ecx
        xor     eax, ebx
        add     edx, 28
        add     edi, eax            ; add function result
        add     ebp, 4e0811a1H      ; add T to next A
        rol     edi, 15             ; rotate
        add     ebp, [edx]          ; add X(13) to next A
        add     edi, esi            ; add B

        mov     eax, esi            ; set EAX now for next iteration
        or      ecx, edi            ; A in EBP, B in EDI, C in ESI, D in EBX
        not     eax
        xor     ecx, esi
        sub     edx, 36
        add     ebp, ecx            ; add function result
        add     ebx, 0f7537e82H     ; add T to next A
        rol     ebp, 21             ; rotate
        add     ebx, [edx]          ; add X(4) to next A
        add     ebp, edi            ; add B

        mov     ecx, edi            ; set ECX now for next iteration
        or      eax, ebp            ; A in EBX, B in EBP, C in EDI, D in ESI
        not     ecx
        xor     eax, edi
        add     edx, 28
        add     ebx, eax            ; add function result
        add     esi, 0bd3af235H     ; add T to next A
        rol     ebx, 6              ; rotate
        add     esi, [edx]          ; add X(11) to next A
        add     ebx, ebp            ; add B

        mov     eax, ebp            ; set EAX now for next iteration
        or      ecx, ebx            ; A in ESI, B in EBX, C in EBP, D in EDI
        not     eax
        xor     ecx, ebp
        sub     edx, 36
        add     esi, ecx            ; add function result
        add     edi, 2ad7d2bbH      ; add T to next A
        rol     esi, 10             ; rotate
        add     edi, [edx]          ; add X(2) to next A
        add     esi, ebx            ; add B

        mov     ecx, ebx            ; set ECX now for next iteration
        or      eax, esi            ; A in EDI, B in ESI, C in EBX, D in EBP
        not     ecx
        xor     eax, ebx
        add     edx, 28
        add     edi, eax            ; add function result
        add     ebp, 0eb86d391H     ; add T to next A
        rol     edi, 15             ; rotate
        add     ebp, [edx]          ; add X(9) to next A
        add     edi, esi            ; add B

        pop     eax
        or      ecx, edi            ; A in EBP, B in EDI, C in ESI, D in EBX
        add     ebx, eax            ; previous MD5State.A value added to final
        xor     ecx, esi
        add     edx, 28             ; EDX increased to next block base value
        add     ebp, ecx            ; add function result
        pop     eax
        rol     ebp, 21             ; rotate
        pop     ecx
        add     ebp, edi            ; add B

        add     edi, ecx            ; previous MD5State.C value added to final
        add     ebp, eax            ; previous MD5State.B value added to final
        pop     eax
        pop     ecx                 ; retrieve counter
        add     esi, eax            ; previous MD5State.D value added to final

        sub     ecx, 64
        jnz     MD5TransformLoop

        pop     eax                 ; retrieve alignment value
        test    eax, eax            ; check alignment - do we need to copy any more data?
        jnz     NonalignedSource
@@:
        ret

NonalignedSource:
        pop     ecx                 ; retrieve master counter
        pop     edx                 ; retrieve original source array pointer
        test    ecx, ecx            ; did we finish processing the source array yet?
        jz      @B

CopyTransformData:
        push    edx                 ; save source array pointer
        push    ecx                 ; save master counter
        push    eax                 ; save alignment value for next loop end as flag

        cmp     ecx, KBytes64       ; how many bytes are we transferring?
        jb      @F
        mov     ecx, KBytes64       ; max size of buffer
@@:
        cld
        sub     [esp+4], ecx        ; subtract this transfer total from the master counter
        push    esi                 ; save MD5State.D
        push    edi                 ; save MD5State.C
        mov     esi, edx            ; set MOVSD source and destination pointers
        mov     edi, MD5Buffer
        mov     eax, ecx            ; save counter for after the byte move operation

        rep     movsd

        mov     ecx, eax            ; retrieve counter
        pop     edi                 ; retrieve MD5State.C
        pop     esi                 ; retrieve MD5State.D
        mov     edx, MD5Buffer      ; set copied source array pointer

        jmp     MD5TransformLoop

MD5Transform    endp


; 
